package edu.utexas.cs.surdules.pipes.model;

import edu.utexas.cs.surdules.pipes.model.io.Visitor;
import edu.utexas.cs.surdules.pipes.view.ConnectorUI;

import java.awt.Point;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

/**
 * @author <A HREF="mailto:razvan.surdulescu@post.harvard.edu">Razvan Surdulescu</A> (c) 2003
 * <BR>
 * You may use and distribute this software for free provided you include
 * this copyright notice. You may not sell this software, use my name for
 * publicity reasons or modify the code without permission from me.
 */
public class Connector {
    public static final int INCOMING = 0;
    public static final int OUTGOING = 1;

    public static final int WEST = 0;
    public static final int SOUTH = 1;
    public static final int EAST = 2;
    public static final int NORTH = 3;

    private Widget m_widget;
    private int m_uid;

    private int m_direction;
    private int m_size;
    private int m_position;

    private int m_x, m_y;

    private boolean m_flip;

    private List m_arcs = new ArrayList();

    private ConnectorUI m_connectorUI;

    public Connector(Widget widget, int direction, int size, int position, int uid) {
        m_widget = widget;

        m_direction = direction;
        m_size = size;
        m_position = position;

        m_uid = uid;

        computeBounds();
    }

    private void computeBounds() {
        int width = m_widget.getWidth();
        int height = m_widget.getHeight();

        switch (m_position) {
            case WEST:
                if (!m_flip) {
                    m_x = 0;
                    m_y = (height - m_size) / 2;
                } else {
                    m_x = width - m_size;
                    m_y = (height - m_size) / 2;
                }
                break;
            case SOUTH:
                if (!m_flip) {
                    m_x = (width - m_size) / 2;
                    m_y = height - m_size;
                } else {
                    m_x = (width - m_size) / 2;
                    m_y = 0;
                }
                break;
            case EAST:
                if (!m_flip) {
                    m_x = width - m_size;
                    m_y = (height - m_size) / 2;
                } else {
                    m_x = 0;
                    m_y = (height - m_size) / 2;
                }
                break;
            case NORTH:
                if (!m_flip) {
                    m_x = (width - m_size) / 2;
                    m_y = 0;
                } else {
                    m_x = (width - m_size) / 2;
                    m_y = height - m_size;
                }
                break;
            default:
                throw new IllegalArgumentException("Invalid Connector position: " + m_position);
        }
    }

    public Widget getWidget() {
        return m_widget;
    }

    public String getUID() {
        return m_widget.getName() + "." + m_uid;
    }

    public int getDirection() {
        return m_direction;
    }

    public void setDirection(int direction) {
        if (direction == INCOMING || direction == OUTGOING) {
            m_direction = direction;

            computeBounds();

            if (m_connectorUI != null) {
                m_connectorUI.touch();
            }
        } else {
            throw new IllegalArgumentException("Invalid Connector direction: " + direction);
        }
    }

    public int getSize() {
        return m_size;
    }

    public void setSize(int size) {
        m_size = size;

        computeBounds();

        if (m_connectorUI != null) {
            m_connectorUI.touch();
        }
    }

    public int getPosition() {
        return m_position;
    }

    public void setPosition(int position) {
        if (position == WEST || position == NORTH || position == EAST || position == SOUTH) {
            m_position = position;

            computeBounds();

            if (m_connectorUI != null) {
                m_connectorUI.touch();
            }
        } else {
            throw new IllegalArgumentException("Invalid Connector position: " + position);
        }
    }

    public int getX() {
        return m_x;
    }

    public int getY() {
        return m_y;
    }

    public Point getCenter() {
        return new Point(m_size / 2, m_size / 2);
    }

    public boolean isFlipped() {
        return m_flip;
    }

    public void flip() {
        m_flip = !m_flip;

        computeBounds();

        if (m_connectorUI != null) {
            m_connectorUI.touch();
        }
    }

    public void update() {
        computeBounds();

        if (m_connectorUI != null) {
            m_connectorUI.touch();
        }
    }

    public void startArc() {
        m_widget.getModel().setStartArc(this);
    }

    public void drawArc(Point point) {
        m_widget.getModel().setDrawArc(point);
    }

    public void endArc(Connector end) {
        m_widget.getModel().setEndArc(end);
    }

    public void finishArc() throws InvalidModelException {
        m_widget.getModel().finishArc();
    }

    public void connect(Connector other) throws InvalidModelException {
        if (m_direction == other.getDirection()) {
            throw new InvalidModelException("Connectors must have different directions");
        }

        if (m_widget == other.getWidget()) {
            throw new InvalidModelException("Connectors must belong to different parents");
        }

        if (hasNeighbor(other) || other.hasNeighbor(this)) {
            throw new InvalidModelException("Connectors already connected");
        }

        addNeighbor(other);
        other.addNeighbor(this);
    }

    public void disconnect(Connector other) throws InvalidModelException {
        if (m_direction == other.getDirection()) {
            throw new InvalidModelException("Connectors must have different directions");
        }

        if (m_widget == other.getWidget()) {
            throw new InvalidModelException("Connectors must belong to different parents");
        }

        if (!hasNeighbor(other) || !other.hasNeighbor(this)) {
            throw new InvalidModelException("Connectors not already connected");
        }

        removeNeighbor(other);
        other.removeNeighbor(this);
    }

    public boolean hasNeighbor(Connector other) {
        for (Iterator iter = m_arcs.iterator(); iter.hasNext();) {
            Arc arc = (Arc) iter.next();
            if (arc.getEnd() == other) {
                return true;
            }
        }

        return false;
    }

    public void addNeighbor(Connector other) throws InvalidModelException {
        if (hasNeighbor(other)) {
            throw new InvalidModelException("Connectors already connected");
        }

        m_arcs.add(new Arc(this, other));
    }

    public void removeNeighbor(Connector other) throws InvalidModelException {
        if (!hasNeighbor(other)) {
            throw new InvalidModelException("Connectors not already connected");
        }

        Arc arc = null;

        for (Iterator iter = m_arcs.iterator(); iter.hasNext();) {
            arc = (Arc) iter.next();
            if (arc.getEnd() == other) {
                break;
            }
        }

        m_arcs.remove(arc);
    }

    public List getArcs() {
        return m_arcs;
    }

    public void setConnectorUI(ConnectorUI connectorUI) {
        m_connectorUI = connectorUI;
    }

    public ConnectorUI getConnectorUI() {
        return m_connectorUI;
    }

    public void accept(Visitor visitor) {
        visitor.visitConnector(this);
        for (Iterator iter = m_arcs.iterator(); iter.hasNext();) {
            Arc arc = (Arc) iter.next();
            arc.accept(visitor);
        }
    }
}
