package edu.utexas.cs.surdules.pipes.model;

import java.util.Random;

/**
 * This package provides the same methods for random drawing as can be found
 * in SIMULA.
 *
 * Taken directly from
 * <A HREF="http://www.dat.ruc.dk/~keld/research/JAVASIMULATION/JAVASIMULATION-1.0/docs/Report.pdf">
 * Discrete Event Simulation in Java</A>.
 *
 * @author <A HREF="mailto:razvan.surdulescu@post.harvard.edu">Razvan Surdulescu</A> (c) 2003
 * <BR>
 * You may use and distribute this software for free provided you include
 * this copyright notice. You may not sell this software, use my name for
 * publicity reasons or modify the code without permission from me.
 */
public class Distributions extends Random {
    public static final String FIXED = "Fixed(a)";
    public static final String UNIFORM = "Uniform(a,b)";
    public static final String NORMAL = "Normal(a,b)";
    public static final String NEGATIVEEXPONENTIAL = "Negative Exponential(a)";
    public static final String POISSON = "Poisson(a)";
    public static final String ERLANG = "Erlang(a,b)";

    private static final String[] s_distributionNames = new String[]{
        FIXED, UNIFORM, NORMAL, NEGATIVEEXPONENTIAL, POISSON, ERLANG
    };

    public static String[] getDistributionNames() {
        return s_distributionNames;
    }

    public double computeNamedDistribution(String name, double a, double b) {
        if (name.equals(FIXED)) {
            return a;
        } else if (name.equals(UNIFORM)) {
            return uniform(a, b);
        } else if (name.equals(NORMAL)) {
            return normal(a, b);
        } else if (name.equals(NEGATIVEEXPONENTIAL)) {
            return negexp(1 / a);
        } else if (name.equals(POISSON)) {
            return poisson(a);
        } else if (name.equals(ERLANG)) {
            return erlang(1 / a, 1 / b);
        } else {
            throw new IllegalArgumentException("Unrecognized distribution: '" + name + "'");
        }
    }

    /**
     * This constructor creates a Random object with the current time as its
     * seed value.
     */
    public Distributions() {
        super();
    }

    /**
     * This constructor creates a Random object with the given seed value.
     */
    public Distributions(long seed) {
        super(seed);
    }

    /**
     * The value is true with the probability a, false with probability 1-a.
     * It is always true if a >= 1, and always false if a <= 0.
     */
    public boolean draw(double a) {
        return a < nextDouble();
    }

    /**
     * The value is one of the integers a, a+1, ..., b-1, b with equal probability.
     * If b < a, the call constitutes an error.
     */
    public int randInt(int a, int b) {
        if (b < a) {
            error("Second parameter is lower than first parameter");
        }
        return (int) (a + nextDouble() * (b - a + 1));
    }

    /**
     * The value is uniformly distributed in the interval a <= x < b. If b <= a,
     * the call constitutes an error.
     */
    public double uniform(double a, double b) {
        if (b <= a) {
            error("Second parameter is not greater than first parameter");
        }
        return a + nextDouble() * (b - a);
    }

    /**
     * The value is normally distributed with mean a and standard deviation b.
     */
    public double normal(double a, double b) {
        return a + b * nextGaussian();
    }

    /**
     * The value is a drawing from the negative exponential distribution with
     * mean 1/a. If a is non-positive, a runtime error occurs.
     */
    public double negexp(double a) {
        if (a <= 0) {
            error("First parameter is lower than zero");
        }
        return -Math.log(nextDouble()) / a;
    }

    /**
     * The value is a drawing from the Poisson distribution with parameter a.
     */
    public int poisson(double a) {
        double limit = Math.exp(-a), prod = nextDouble();
        int n;
        for (n = 0; prod >= limit; n++) {
            prod *= nextDouble();
        }
        return n;
    }

    /**
     * The value is a drawing from the Erlang distribution with mean 1/a
     * and standard deviation 1/(a* sqrt(b)). Both a and b must be positive.
     */
    public double erlang(double a, double b) {
        if (a <= 0) {
            error("First parameter is not greater than zero");
        }
        if (b <= 0) {
            error("Second parameter is not greater than zero");
        }
        long bi = (long) b, ci;
        if (bi == b) {
            bi--;
        }
        double sum = 0;
        for (ci = 1; ci <= bi; ci++) {
            sum += Math.log(nextDouble());
        }
        return -(sum + (b - (ci - 1)) * Math.log(nextDouble())) /
                (a * b);
    }

    /**
     * The one-dimensional array a of n elements of type double, augmented
     * by the element 1 to the right, is interpreted as a step function
     * of the subscript, defining a discrete (cumulative) distribution function.
     * The function value satisfies
     *
     * 0 <= discrete(a) <= n
     *
     * It is defined as the smallest i such that a[i] > r, where r is a random
     * number in the interval [0;1] and a[n] = 1.
     */
    public int discrete(double[] a) {
        double basic = nextDouble();
        int i;
        for (i = 0; i < a.length; i++) {
            if (a[i] > basic) {
                break;
            }
        }
        return i;
    }

    /**
     * The value is a drawing from a (cumulative) distribution function f,
     * which is obtained by linear interpolation in a non-equidistant table defined
     * by a and b, such that a[i] = f(b[i]).
     *
     * It is assumed that a and b are one-dimensional arrays of the same
     * length, that the first and last elements of a are equal to 0 and 1, respectively,
     * and that a[i] >= a[j] and b[i] > b[j] for i > j.
     */
    public double linear(double[] a, double[] b) {
        if (a.length != b.length) {
            error("Arrays have different length");
        }
        if (a[0] != 0.0 || a[a.length - 1] != 1.0) {
            error("Illegal value in first array");
        }
        double basic = nextDouble();
        int i;
        for (i = 1; i < a.length; i++) {
            if (a[i] >= basic) {
                break;
            }
        }
        double d = a[i] - a[i - 1];
        if (d == 0.0) {
            return b[i - 1];
        }
        return b[i - 1] + (b[i] - b[i - 1]) * (basic - a[i - 1]) / d;
    }

    /**
     * The value is an integer in the range [0;n-1] where n is the number of
     * elements in the one-dimensional array a. The latter is interpreted as a
     * histogram defining the relative frequencies of the values.
     */
    public int histd(double[] a) {
        double sum = 0.0;
        int i;
        for (i = 0; i < a.length; i++) {
            sum += a[i];
        }
        double weight = nextDouble() * sum;
        sum = 0.0;
        for (i = 0; i < a.length - 1; i++) {
            sum += a[i];
            if (sum >= weight) {
                break;
            }
        }
        return i;
    }

    private void error(String msg) {
        throw new IllegalArgumentException(msg);
    }
}
