package edu.utexas.cs.surdules.pipes.model;

import edu.utexas.cs.surdules.pipes.model.forms.PropertiesItem;
import edu.utexas.cs.surdules.pipes.model.forms.PropertiesList;
import edu.utexas.cs.surdules.pipes.model.io.Visitor;
import edu.utexas.cs.surdules.pipes.view.ModelUI;

import java.awt.Point;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

/**
 * @author <A HREF="mailto:razvan.surdulescu@post.harvard.edu">Razvan Surdulescu</A> (c) 2003
 * <BR>
 * You may use and distribute this software for free provided you include
 * this copyright notice. You may not sell this software, use my name for
 * publicity reasons or modify the code without permission from me.
 */
public class Model {
    private static int s_globalCount = 0;

    private ModelUI m_modelUI;

    private Connector m_startArc;
    private Point m_drawArc;
    private Connector m_endArc;

    private Arc m_selectedArc;

    private String m_name;
    private double m_warmupTime;
    private double m_simulationTime;

    private List m_widgets = new ArrayList();

    public Model() {
        s_globalCount++;
        m_name = "Model" + s_globalCount;
    }

    public void addWidget(Widget widget) throws InvalidModelException {
        if (m_widgets.contains(widget)) {
            throw new InvalidModelException("Widget already exists");
        }
        m_widgets.add(widget);

        if (m_modelUI != null) {
            m_modelUI.touch();
        }
    }

    public List getWidgets() {
        return m_widgets;
    }

    public List getWidgets(Class type) {
        List widgets = new ArrayList();
        for (Iterator iter = m_widgets.iterator(); iter.hasNext();) {
            Widget widget = (Widget) iter.next();
            if (widget.getClass().isAssignableFrom(type)) {
                widgets.add(widget);
            }
        }
        return widgets;
    }

    public Widget getWidget(String name) {
        for (Iterator iter = m_widgets.iterator(); iter.hasNext();) {
            Widget widget = (Widget) iter.next();
            if (widget.getName().equals(name)) {
                return widget;
            }
        }

        return null;
    }

    public void setStartArc(Connector connector) {
        m_startArc = connector;
    }

    public Connector getStartArc() {
        return m_startArc;
    }

    public void setDrawArc(Point point) {
        m_drawArc = point;

        if (m_modelUI != null) {
            m_modelUI.repaint();
        }
    }

    public Point getDrawArc() {
        return m_drawArc;
    }

    public void setEndArc(Connector endArc) {
        m_endArc = endArc;
    }

    public Connector getEndArc() {
        return m_endArc;
    }

    public void finishArc() throws InvalidModelException {
        try {
            if (m_startArc == null || m_endArc == null) {
                throw new InvalidModelException("Invalid start or end point");
            }

            m_startArc.connect(m_endArc);
        } finally {
            m_startArc = null;
            m_drawArc = null;
            m_endArc = null;

            if (m_modelUI != null) {
                m_modelUI.repaint();
            }
        }
    }

    public void deleteWidget(Widget widget) throws InvalidModelException {
        if (!m_widgets.contains(widget)) {
            throw new InvalidModelException("Widget not found");
        }

        m_widgets.remove(widget);

        if (m_modelUI != null) {
            m_modelUI.touch();
        }
    }

    public boolean selectArc(Point point) {
        for (Iterator widgets = m_widgets.iterator(); widgets.hasNext();) {
            Widget widget = (Widget) widgets.next();

            for (Iterator connectors = widget.getConnectors().iterator(); connectors.hasNext();) {
                Connector connector = (Connector) connectors.next();
                if (connector.getDirection() == Connector.INCOMING) {
                    continue;
                }

                for (Iterator arcs = connector.getArcs().iterator(); arcs.hasNext();) {
                    Arc arc = (Arc) arcs.next();

                    if (arc.contains(point)) {
                        m_selectedArc = arc;
                        return true;
                    }
                }
            }
        }

        return false;
    }

    public Arc getSelectedArc() {
        return m_selectedArc;
    }

    public void deleteSelectedArc() throws InvalidModelException {
        if (m_selectedArc == null) {
            throw new InvalidModelException("No arc selected");
        }

        m_selectedArc.getStart().disconnect(m_selectedArc.getEnd());
        m_selectedArc = null;

        if (m_modelUI != null) {
            m_modelUI.repaint();
        }
    }

    public PropertiesList getPropertiesList() {
        PropertiesList propertiesList = new PropertiesList();

        propertiesList.addItem(new PropertiesItem("Name", String.class, m_name));
        propertiesList.addItem(new PropertiesItem("Warmup Time", Double.class, new Double(m_warmupTime)));
        propertiesList.addItem(new PropertiesItem("Simulation Time", Double.class, new Double(m_simulationTime)));

        return propertiesList;
    }

    public void setPropertiesList(PropertiesList propertiesList) {
        setName((String) propertiesList.getItem("Name").getValue());

        m_warmupTime = ((Double) propertiesList.getItem("Warmup Time").getValue()).doubleValue();
        m_simulationTime = ((Double) propertiesList.getItem("Simulation Time").getValue()).doubleValue();
    }

    public String getName() {
        return m_name;
    }

    public void setName(String name) {
        m_name = name;

        if (m_modelUI != null) {
            m_modelUI.touch();
        }
    }

    public double getWarmupTime() {
        return m_warmupTime;
    }

    public void setWarmupTime(double warmupTime) {
        if (warmupTime < 0.0) {
            throw new IllegalArgumentException("Warmup time must be positive: " + warmupTime);
        }
        m_warmupTime = warmupTime;
    }

    public double getSimulationTime() {
        return m_simulationTime;
    }

    public void setSimulationTime(double simulationTime) {
        if (simulationTime < 0.0) {
            throw new IllegalArgumentException("Simulation time must be positive: " + simulationTime);
        }
        m_simulationTime = simulationTime;
    }

    public void setModelUI(ModelUI modelUI) {
        m_modelUI = modelUI;
    }

    public ModelUI getModelUI() {
        return m_modelUI;
    }

    public void accept(Visitor visitor) {
        visitor.visitModel(this);
        for (Iterator iter = m_widgets.iterator(); iter.hasNext();) {
            Widget widget = (Widget) iter.next();
            widget.accept(visitor);
        }
    }

    public void startStatistics() {
        for (Iterator iter = m_widgets.iterator(); iter.hasNext();) {
            Widget widget = (Widget) iter.next();
            widget.getStatistics().onStart();
        }
    }

    public void endStatistics(double time) {
        for (Iterator iter = m_widgets.iterator(); iter.hasNext();) {
            Widget widget = (Widget) iter.next();
            widget.getStatistics().onEnd(time);
        }
    }
}
