package edu.utexas.cs.surdules.pipes.model;

import edu.utexas.cs.surdules.pipes.model.forms.PropertiesItem;
import edu.utexas.cs.surdules.pipes.model.forms.PropertiesList;
import edu.utexas.cs.surdules.pipes.model.io.Visitor;
import edu.utexas.cs.surdules.pipes.model.simulate.Simulator;
import edu.utexas.cs.surdules.pipes.model.simulate.Event;
import edu.utexas.cs.surdules.pipes.model.statistics.Statistics;
import edu.utexas.cs.surdules.pipes.view.WidgetUI;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;

/**
 * @author <A HREF="mailto:razvan.surdulescu@post.harvard.edu">Razvan Surdulescu</A> (c) 2003
 * <BR>
 * You may use and distribute this software for free provided you include
 * this copyright notice. You may not sell this software, use my name for
 * publicity reasons or modify the code without permission from me.
 */
public abstract class Widget {
    private static long s_globalCount = 0;

    protected Model m_model;

    private int m_x, m_y;
    private int m_width, m_height;

    private String m_name;

    private boolean m_flip;

    protected List m_connectors = new ArrayList();

    protected Statistics m_statistics = new Statistics(this);

    private WidgetUI m_widgetUI;

    public Widget(Model model) {
        this(model, 0, 0);
    }

    public Widget(Model model, int x, int y) {
        m_model = model;
        m_x = x;
        m_y = y;

        s_globalCount++;
        if (s_globalCount == 0) {
            // we rolled over: this should never happen
            throw new Error("Widget global count rolled over");
        }

        m_name = getDefaultName() + s_globalCount;

        createConnectors();
    }

    protected abstract void createConnectors();

    protected abstract String getDefaultName();

    public abstract Event advance(Event event, Simulator context);

    public PropertiesList getPropertiesList() {
        PropertiesList propertiesList = new PropertiesList();
        propertiesList.addItem(new PropertiesItem("Name", String.class, m_name));
        return propertiesList;
    }

    public void setPropertiesList(PropertiesList propertiesList) {
        setName((String) propertiesList.getItem("Name").getValue());
    }

    public Statistics getStatisticsCopy() {
        return new Statistics(m_statistics);
    }

    public Statistics getStatistics() {
        return m_statistics;
    }

    public void setStatistics(Statistics statistics) {
        m_statistics = statistics;
    }

    public Model getModel() {
        return m_model;
    }

    public void setCoordinates(int x, int y) {
        m_x = x;
        m_y = y;

        if (m_widgetUI != null) {
            m_widgetUI.touch();
        }
    }

    public void setX(int x) {
        m_x = x;

        if (m_widgetUI != null) {
            m_widgetUI.touch();
        }
    }

    public void setY(int y) {
        m_y = y;

        if (m_widgetUI != null) {
            m_widgetUI.touch();
        }
    }

    public int getX() {
        return m_x;
    }

    public int getY() {
        return m_y;
    }

    public void setDimensions(int width, int height) {
        m_width = width;
        m_height = height;

        for (Iterator iter = m_connectors.iterator(); iter.hasNext();) {
            Connector connector = (Connector) iter.next();
            connector.update();
        }

        if (m_widgetUI != null) {
            m_widgetUI.touch();
        }
    }

    public int getWidth() {
        return m_width;
    }

    public int getHeight() {
        return m_height;
    }

    public boolean isFlipped() {
        return m_flip;
    }

    public void flip() {
        m_flip = !m_flip;

        for (Iterator iter = m_connectors.iterator(); iter.hasNext();) {
            Connector connector = (Connector) iter.next();
            connector.flip();
        }

        if (m_widgetUI != null) {
            m_widgetUI.touch();
        }
    }

    public void delete() {
        try {
            for (Iterator connectors = m_connectors.iterator(); connectors.hasNext();) {
                Connector connector = (Connector) connectors.next();

                HashSet copy = new HashSet(connector.getArcs());
                for (Iterator arcs = copy.iterator(); arcs.hasNext();) {
                    Arc arc = (Arc) arcs.next();
                    connector.disconnect(arc.getEnd());
                }
            }

            m_model.deleteWidget(this);
        } catch (InvalidModelException ime) {
            // this should never happen
            throw new RuntimeException("Corrupted model: " + ime.getMessage());
        }
    }

    public List getConnectors() {
        return m_connectors;
    }

    // TODO: assumes only one OUTGOING connector
    public Arc selectOutgoingArc(String category, Simulator context) {
        for (Iterator connectors = m_connectors.iterator(); connectors.hasNext();) {
            Connector connector = (Connector) connectors.next();
            if (connector.getDirection() == Connector.INCOMING) {
                continue;
            }

            double draw = context.getDistributions().uniform(0, 1);
            double low = 0;

            for (Iterator arcs = connector.getArcs().iterator(); arcs.hasNext();) {
                Arc arc = (Arc) arcs.next();
                if (arc.getSelectCategory() == null || arc.getSelectCategory().equals(category)) {
                    if (low <= draw && draw < low + arc.getSelectProbability()) {
                        return arc;
                    } else {
                        low = arc.getSelectProbability();
                    }
                }
            }
        }

        return null;
    }

    public String getName() {
        return m_name;
    }

    public void setName(String name) {
        if (name.equals("")) {
            throw new IllegalArgumentException("Empty name");
        }

        Widget other = m_model.getWidget(name);
        if (other != null && other != this) {
            throw new IllegalArgumentException("Duplicate name: '" + name + "'");
        }

        setUniqueName(name);
    }

    public void setUniqueName(String name) {
        m_name = name;

        if (m_widgetUI != null) {
            m_widgetUI.touch();
        }
    }

    public void setWidgetUI(WidgetUI widgetUI) {
        m_widgetUI = widgetUI;
    }

    public WidgetUI getWidgetUI() {
        return m_widgetUI;
    }

    public void accept(Visitor visitor) {
        for (Iterator iter = m_connectors.iterator(); iter.hasNext();) {
            Connector connector = (Connector) iter.next();
            connector.accept(visitor);
        }
    }
}
