package edu.utexas.cs.surdules.pipes.model.io;

import edu.utexas.cs.surdules.pipes.model.Arc;
import edu.utexas.cs.surdules.pipes.model.Connector;
import edu.utexas.cs.surdules.pipes.model.InvalidModelException;
import edu.utexas.cs.surdules.pipes.model.Model;
import edu.utexas.cs.surdules.pipes.model.Widget;
import edu.utexas.cs.surdules.pipes.model.widgets.WidgetAllocate;
import edu.utexas.cs.surdules.pipes.model.widgets.WidgetDelay;
import edu.utexas.cs.surdules.pipes.model.widgets.WidgetRelease;
import edu.utexas.cs.surdules.pipes.model.widgets.WidgetResource;
import edu.utexas.cs.surdules.pipes.model.widgets.WidgetService;
import edu.utexas.cs.surdules.pipes.model.widgets.WidgetSink;
import edu.utexas.cs.surdules.pipes.model.widgets.WidgetSource;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;
import org.xml.sax.SAXException;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

/**
 * @author <A HREF="mailto:razvan.surdulescu@post.harvard.edu">Razvan Surdulescu</A> (c) 2003
 * <BR>
 * You may use and distribute this software for free provided you include
 * this copyright notice. You may not sell this software, use my name for
 * publicity reasons or modify the code without permission from me.
 */
public class LoadVisitor implements Visitor {
    private String m_fileName;

    private Document m_document;
    private Element m_root;

    private Model m_currentModel;

    private Map m_widgetElements = new HashMap();
    private Map m_connectorElements = new HashMap();
    private Map m_arcElements = new HashMap();

    public LoadVisitor(String fileName) {
        if (fileName == null) {
            throw new IllegalArgumentException("File name cannot be null");
        }

        m_fileName = fileName;
    }

    private String readAttribute(Element element, String name) {
        return element.getAttribute(name);
    }

    private String readChild(Element parent, String name) {
        NodeList children = parent.getElementsByTagName(name);

        if (children.getLength() != 1) {
            throw new RuntimeException("Corrupted model: invalid number of elements of type '" + name + "'");
        }

        Node childNode = children.item(0);
        if (childNode.getNodeType() == Node.ELEMENT_NODE) {
            Element childElement = (Element) childNode;
            Node textNode = childElement.getFirstChild();
            if (textNode == null) {
                return "";
            } else if (textNode.getNodeType() == Node.TEXT_NODE) {
                return ((Text) textNode).getData();
            }
        }

        throw new RuntimeException("Corrupted model: no TEXT node found under '" + parent.getNodeName() + "." + name + "'");
    }

    public void visitModel(Model model) {
        m_currentModel = model;

        model.setName(m_root.getAttribute("name"));
        model.setWarmupTime(Double.parseDouble(m_root.getAttribute("warmup")));
        model.setSimulationTime(Double.parseDouble(m_root.getAttribute("simulation")));

        NodeList widgetNodes = m_root.getChildNodes();
        for (int i = 0; i < widgetNodes.getLength(); i++) {
            Node widgetNode = widgetNodes.item(i);
            if (widgetNode.getNodeType() == Node.ELEMENT_NODE) {
                Element widgetElement = (Element) widgetNode;
                if (widgetElement.getNodeName().startsWith("widget")) {
                    Widget widget = XMLFactory.createWidgetForXMLType(widgetElement.getAttribute("type"), model);
                    widget.setUniqueName(widgetElement.getAttribute("name"));

                    try {
                        model.addWidget(widget);
                    } catch (InvalidModelException ime) {
                        throw new RuntimeException("Corrupted model: " + ime.getMessage());
                    }

                    m_widgetElements.put(widgetElement.getAttribute("name"), widgetElement);

                    NodeList connectorNodes = widgetElement.getChildNodes();
                    for (int j = 0; j < connectorNodes.getLength(); j++) {
                        Node connectorNode = connectorNodes.item(j);
                        if (connectorNode.getNodeType() == Node.ELEMENT_NODE) {
                            Element connectorElement = (Element) connectorNode;

                            if (connectorElement.getNodeName().equals("connector")) {
                                m_connectorElements.put(connectorElement.getAttribute("uid"), connectorElement);

                                NodeList arcNodes = connectorElement.getChildNodes();
                                for (int k = 0; k < arcNodes.getLength(); k++) {
                                    Node arcNode = arcNodes.item(k);
                                    if (arcNode.getNodeType() == Node.ELEMENT_NODE) {
                                        Element arcElement = (Element) arcNode;

                                        if (arcElement.getNodeName().equals("arc")) {
                                            m_arcElements.put(connectorElement.getAttribute("uid") + arcElement.getAttribute("end"), arcElement);
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    private void visitWidget(Widget widget) {
        Element currentWidgetElement = (Element) m_widgetElements.get(widget.getName());
        widget.setX(Integer.parseInt(currentWidgetElement.getAttribute("x")));
        widget.setY(Integer.parseInt(currentWidgetElement.getAttribute("y")));
        if (Boolean.getBoolean(currentWidgetElement.getAttribute("flip"))) {
            widget.flip();
        }

        NodeList children = currentWidgetElement.getElementsByTagName("statlet");
        for (int i = 0; i < children.getLength(); i++) {
            Node childNode = children.item(i);
            if (childNode.getNodeType() == Node.ELEMENT_NODE) {
                Element childElement = (Element) childNode;
                Node textNode = childElement.getFirstChild();
                if (textNode.getNodeType() == Node.TEXT_NODE) {
                    widget.getStatistics().enableStatlet(((Text) textNode).getData());
                }
            }
        }
    }

    public void visitWidgetAllocate(WidgetAllocate widget) {
        visitWidget(widget);

        Element currentWidgetElement = (Element) m_widgetElements.get(widget.getName());

        String resourceName = readChild(currentWidgetElement, "resource");
        if (!resourceName.equals("")) {
            widget.setResource((WidgetResource) m_currentModel.getWidget(resourceName));
        }

        widget.setDistributionName(readChild(currentWidgetElement, "distribution-name"));
        widget.setDistributionValueA(Double.parseDouble(readChild(currentWidgetElement, "distribution-value-a")));
        widget.setDistributionValueB(Double.parseDouble(readChild(currentWidgetElement, "distribution-value-b")));
        widget.setTimeRuleName(readChild(currentWidgetElement, "time-rule"));
        widget.setQueueingPriorityName(readChild(currentWidgetElement, "queueing-priority"));
    }

    public void visitWidgetDelay(WidgetDelay widget) {
        visitWidget(widget);

        Element currentWidgetElement = (Element) m_widgetElements.get(widget.getName());
        widget.setDistributionName(readChild(currentWidgetElement, "distribution-name"));
        widget.setDistributionValueA(Double.parseDouble(readChild(currentWidgetElement, "distribution-value-a")));
        widget.setDistributionValueB(Double.parseDouble(readChild(currentWidgetElement, "distribution-value-b")));
    }

    public void visitWidgetRelease(WidgetRelease widget) {
        visitWidget(widget);

        Element currentWidgetElement = (Element) m_widgetElements.get(widget.getName());

        String resourceName = readChild(currentWidgetElement, "resource");
        if (!resourceName.equals("")) {
            widget.setResource((WidgetResource) m_currentModel.getWidget(resourceName));
        }

        widget.setDistributionName(readChild(currentWidgetElement, "distribution-name"));
        widget.setDistributionValueA(Double.parseDouble(readChild(currentWidgetElement, "distribution-value-a")));
        widget.setDistributionValueB(Double.parseDouble(readChild(currentWidgetElement, "distribution-value-b")));
    }

    public void visitWidgetResource(WidgetResource widget) {
        visitWidget(widget);

        Element currentWidgetElement = (Element) m_widgetElements.get(widget.getName());
        widget.setNumberOfTokens(Integer.parseInt(readChild(currentWidgetElement, "tokens")));
    }

    public void visitWidgetService(WidgetService widget) {
        visitWidget(widget);

        Element currentWidgetElement = (Element) m_widgetElements.get(widget.getName());
        widget.setDistributionName(readChild(currentWidgetElement, "distribution-name"));
        widget.setDistributionValueA(Double.parseDouble(readChild(currentWidgetElement, "distribution-value-a")));
        widget.setDistributionValueB(Double.parseDouble(readChild(currentWidgetElement, "distribution-value-b")));
        widget.setTimeRuleName(readChild(currentWidgetElement, "time-rule"));
        widget.setQueueingPriorityName(readChild(currentWidgetElement, "queueing-priority"));
    }

    public void visitWidgetSink(WidgetSink widget) {
        visitWidget(widget);
    }

    public void visitWidgetSource(WidgetSource widget) {
        visitWidget(widget);

        Element currentWidgetElement = (Element) m_widgetElements.get(widget.getName());
        widget.setDistributionName(readChild(currentWidgetElement, "distribution-name"));
        widget.setDistributionValueA(Double.parseDouble(readChild(currentWidgetElement, "distribution-value-a")));
        widget.setDistributionValueB(Double.parseDouble(readChild(currentWidgetElement, "distribution-value-b")));
        widget.setNumberOfTransactions(Integer.parseInt(readChild(currentWidgetElement, "transactions")));

        String category = readChild(currentWidgetElement, "category");
        if (!category.equals("")) {
            widget.setTransactionCategory(category);
        }

        widget.setTransactionPriority(Integer.parseInt(readChild(currentWidgetElement, "priority")));
    }

    public void visitConnector(Connector connector) {
        Element currentConnectorElement = (Element) m_connectorElements.get(connector.getUID());

        connector.setDirection(Integer.parseInt(currentConnectorElement.getAttribute("direction")));
        connector.setSize(Integer.parseInt(currentConnectorElement.getAttribute("size")));
        connector.setPosition(Integer.parseInt(currentConnectorElement.getAttribute("position")));

        NodeList arcNodes = currentConnectorElement.getChildNodes();
        for (int i = 0; i < arcNodes.getLength(); i++) {
            Node arcNode = arcNodes.item(i);
            if (arcNode.getNodeType() == Node.ELEMENT_NODE) {
                Element arcElement = (Element) arcNode;

                if (arcElement.getNodeName().equals("arc")) {
                    String endUID = arcElement.getAttribute("end");

                    boolean done = false;
                    for (Iterator widgets = connector.getWidget().getModel().getWidgets().iterator(); widgets.hasNext() && !done;) {
                        Widget widget = (Widget) widgets.next();

                        for (Iterator connectors = widget.getConnectors().iterator(); connectors.hasNext() && !done;) {
                            Connector end = (Connector) connectors.next();
                            if (end.getUID().equals(endUID)) {
                                try {
                                    connector.connect(end);
                                } catch (InvalidModelException ime) {
                                    // this should never happen
                                    throw new RuntimeException("Corrupted model: " + ime.getMessage());
                                }
                                done = true;
                            }
                        }
                    }
                }
            }
        }
    }

    public void visitArc(Arc arc) {
        if (arc.getStart().getDirection() == Connector.INCOMING) {
            return;
        }

        Element arcElement = (Element) m_arcElements.get(arc.getStart().getUID() + arc.getEnd().getUID());

        arc.setSelectProbability(Double.parseDouble(arcElement.getAttribute("probability")));

        String category = arcElement.getAttribute("category");
        if (!category.equals("")) {
            arc.setSelectCategory(category);
        }
    }

    public void load() throws IOException, SAXException {
        m_document = XMLFactory.loadDocument(new File(m_fileName));

        m_root = m_document.getDocumentElement();
    }
}
