package edu.utexas.cs.surdules.pipes.model.io;

import edu.utexas.cs.surdules.pipes.model.Arc;
import edu.utexas.cs.surdules.pipes.model.Connector;
import edu.utexas.cs.surdules.pipes.model.Model;
import edu.utexas.cs.surdules.pipes.model.Widget;
import edu.utexas.cs.surdules.pipes.model.widgets.WidgetAllocate;
import edu.utexas.cs.surdules.pipes.model.widgets.WidgetDelay;
import edu.utexas.cs.surdules.pipes.model.widgets.WidgetRelease;
import edu.utexas.cs.surdules.pipes.model.widgets.WidgetResource;
import edu.utexas.cs.surdules.pipes.model.widgets.WidgetService;
import edu.utexas.cs.surdules.pipes.model.widgets.WidgetSink;
import edu.utexas.cs.surdules.pipes.model.widgets.WidgetSource;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

import java.io.File;
import java.io.IOException;
import java.util.Iterator;

/**
 * @author <A HREF="mailto:razvan.surdulescu@post.harvard.edu">Razvan Surdulescu</A> (c) 2003
 * <BR>
 * You may use and distribute this software for free provided you include
 * this copyright notice. You may not sell this software, use my name for
 * publicity reasons or modify the code without permission from me.
 */
public class SaveVisitor implements Visitor {
    private String m_fileName;

    private Document m_document;
    private Element m_root;

    private Element m_currentWidgetElement;
    private Element m_currentConnectorElement;

    public SaveVisitor(String fileName) {
        if (fileName == null) {
            throw new IllegalArgumentException("File name cannot be null");
        }

        m_fileName = fileName;

        m_document = XMLFactory.createDocument();
    }

    private void writeChild(Element parent, String name, String value) {
        NodeList children = parent.getElementsByTagName(name);

        if (children.getLength() != 0) {
            // this should never happen
            throw new RuntimeException("Corrupted model: more than one child of type '" + name + "'");
        }

        Element child = m_document.createElement(name);
        parent.appendChild(child);
        Text childText = m_document.createTextNode(value);
        child.appendChild(childText);
    }

    public void visitModel(Model model) {
        m_root = m_document.createElement("model");
        m_document.appendChild(m_root);

        m_root.setAttribute("name", model.getName());
        m_root.setAttribute("warmup", String.valueOf(model.getWarmupTime()));
        m_root.setAttribute("simulation", String.valueOf(model.getSimulationTime()));
    }

    private void visitWidget(Widget widget) {
        m_currentWidgetElement = m_document.createElement("widget");
        m_root.appendChild(m_currentWidgetElement);

        m_currentWidgetElement.setAttribute("type", XMLFactory.createXMLTypeForWidget(widget));
        m_currentWidgetElement.setAttribute("name", widget.getName());
        m_currentWidgetElement.setAttribute("x", String.valueOf(widget.getX()));
        m_currentWidgetElement.setAttribute("y", String.valueOf(widget.getY()));
        m_currentWidgetElement.setAttribute("flip", String.valueOf(widget.isFlipped()));

        for (Iterator iter = widget.getStatistics().getStatlets().keySet().iterator(); iter.hasNext();) {
            String key = (String) iter.next();

            Element statletElement = m_document.createElement("statlet");
            m_currentWidgetElement.appendChild(statletElement);
            Text statletText = m_document.createTextNode(key);
            statletElement.appendChild(statletText);
        }
    }

    public void visitWidgetAllocate(WidgetAllocate widget) {
        visitWidget(widget);

        WidgetResource resource = widget.getResource();
        writeChild(m_currentWidgetElement, "resource", String.valueOf(resource == null ? "" : resource.getName()));
        writeChild(m_currentWidgetElement, "distribution-name", String.valueOf(widget.getDistributionName()));
        writeChild(m_currentWidgetElement, "distribution-value-a", String.valueOf(widget.getDistributionValueA()));
        writeChild(m_currentWidgetElement, "distribution-value-b", String.valueOf(widget.getDistributionValueB()));
        writeChild(m_currentWidgetElement, "time-rule", widget.getTimeRuleName());
        writeChild(m_currentWidgetElement, "queueing-priority", widget.getQueueingPriorityName());
    }

    public void visitWidgetDelay(WidgetDelay widget) {
        visitWidget(widget);

        writeChild(m_currentWidgetElement, "distribution-name", String.valueOf(widget.getDistributionName()));
        writeChild(m_currentWidgetElement, "distribution-value-a", String.valueOf(widget.getDistributionValueA()));
        writeChild(m_currentWidgetElement, "distribution-value-b", String.valueOf(widget.getDistributionValueB()));
    }

    public void visitWidgetRelease(WidgetRelease widget) {
        visitWidget(widget);

        WidgetResource resource = widget.getResource();
        writeChild(m_currentWidgetElement, "resource", String.valueOf(resource == null ? "" : resource.getName()));
        writeChild(m_currentWidgetElement, "distribution-name", String.valueOf(widget.getDistributionName()));
        writeChild(m_currentWidgetElement, "distribution-value-a", String.valueOf(widget.getDistributionValueA()));
        writeChild(m_currentWidgetElement, "distribution-value-b", String.valueOf(widget.getDistributionValueB()));
    }

    public void visitWidgetResource(WidgetResource widget) {
        visitWidget(widget);

        writeChild(m_currentWidgetElement, "tokens", String.valueOf(widget.getNumberOfTokens()));
    }

    public void visitWidgetService(WidgetService widget) {
        visitWidget(widget);

        writeChild(m_currentWidgetElement, "distribution-name", String.valueOf(widget.getDistributionName()));
        writeChild(m_currentWidgetElement, "distribution-value-a", String.valueOf(widget.getDistributionValueA()));
        writeChild(m_currentWidgetElement, "distribution-value-b", String.valueOf(widget.getDistributionValueB()));
        writeChild(m_currentWidgetElement, "time-rule", widget.getTimeRuleName());
        writeChild(m_currentWidgetElement, "queueing-priority", widget.getQueueingPriorityName());
    }

    public void visitWidgetSink(WidgetSink widget) {
        visitWidget(widget);
    }

    public void visitWidgetSource(WidgetSource widget) {
        visitWidget(widget);

        writeChild(m_currentWidgetElement, "distribution-name", String.valueOf(widget.getDistributionName()));
        writeChild(m_currentWidgetElement, "distribution-value-a", String.valueOf(widget.getDistributionValueA()));
        writeChild(m_currentWidgetElement, "distribution-value-b", String.valueOf(widget.getDistributionValueB()));
        writeChild(m_currentWidgetElement, "transactions", String.valueOf(widget.getNumberOfTransactions()));
        writeChild(m_currentWidgetElement, "category",
                String.valueOf(widget.getTransactionCategory() == null ? "" : widget.getTransactionCategory()));
        writeChild(m_currentWidgetElement, "priority", String.valueOf(widget.getTransactionPriority()));
    }

    public void visitConnector(Connector connector) {
        m_currentConnectorElement = m_document.createElement("connector");
        m_currentWidgetElement.appendChild(m_currentConnectorElement);

        m_currentConnectorElement.setAttribute("uid", connector.getUID());
        m_currentConnectorElement.setAttribute("direction", String.valueOf(connector.getDirection()));
        m_currentConnectorElement.setAttribute("size", String.valueOf(connector.getSize()));
        m_currentConnectorElement.setAttribute("position", String.valueOf(connector.getPosition()));
    }

    public void visitArc(Arc arc) {
        if (arc.getStart().getDirection() == Connector.INCOMING) {
            return;
        }

        Element arcElement = m_document.createElement("arc");
        m_currentConnectorElement.appendChild(arcElement);

        arcElement.setAttribute("end", arc.getEnd().getUID());
        arcElement.setAttribute("probability", String.valueOf(arc.getSelectProbability()));
        arcElement.setAttribute("category", (arc.getSelectCategory() == null ? "" : arc.getSelectCategory()));
    }

    public void save() throws IOException {
        XMLFactory.saveDocument(m_document, new File(m_fileName));
    }
}
