package edu.utexas.cs.surdules.pipes.model.simulate;

import edu.utexas.cs.surdules.pipes.model.Distributions;
import edu.utexas.cs.surdules.pipes.model.Model;
import edu.utexas.cs.surdules.pipes.model.QueueingWidget;
import edu.utexas.cs.surdules.pipes.model.Widget;
import edu.utexas.cs.surdules.pipes.model.widgets.WidgetResource;
import edu.utexas.cs.surdules.pipes.model.widgets.WidgetSource;
import edu.utexas.cs.surdules.pipes.view.simulate.SimulatorUI;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

/**
 * @author <A HREF="mailto:razvan.surdulescu@post.harvard.edu">Razvan Surdulescu</A> (c) 2003
 * <BR>
 * You may use and distribute this software for free provided you include
 * this copyright notice. You may not sell this software, use my name for
 * publicity reasons or modify the code without permission from me.
 */
public class Simulator {
    private Model m_model;

    private double m_end;

    private Distributions m_distributions = new Distributions(5);

    private SimulatorUI m_simulatorUI;

    private Set m_events = new TreeSet(new Comparator() {
        public int compare(Object o1, Object o2) {
            Event t1 = (Event) o1;
            Event t2 = (Event) o2;

            if (t1.getTime() < t2.getTime()) {
                return -1;
            } else if (t1.getTime() == t2.getTime()) {
                return (t1.getTransactionID() - t2.getTransactionID());
            } else {
                return 1;
            }
        }
    });

    private Map m_resources = new HashMap();
    private Map m_sources = new HashMap();
    private Map m_queues = new HashMap();

    public Simulator(Model model) {
        m_model = model;
    }

    public Model getModel() {
        return m_model;
    }

    public Distributions getDistributions() {
        return m_distributions;
    }

    public boolean isFinished() {
        return m_events.isEmpty();
    }

    public void start() {
        List m_sources = m_model.getWidgets(WidgetSource.class);
        if (m_sources.size() == 0) {
            return;
        }

        m_model.startStatistics();

        int transactionCount = 0;
        for (Iterator iter = m_sources.iterator(); iter.hasNext();) {
            WidgetSource widgetSource = (WidgetSource) iter.next();

            for (int i = 0; i < widgetSource.getNumberOfTransactions(); i++) {
                Event current = new Event(this, widgetSource, transactionCount++,
                        widgetSource.getTransactionCategory(), widgetSource.getTransactionPriority());
                m_events.add(current);
            }
        }
    }

    private Event pop() {
        TreeSet events = (TreeSet) m_events;
        Event first = (Event) events.first();
        events.remove(first);
        return first;
    }

    public void step() {
        if (m_events.isEmpty()) {
            throw new IllegalStateException("Empty queue");
        }

        Event current = pop();

        if (m_model.getSimulationTime() > 0 && current.getTime() > m_model.getSimulationTime()) {
            m_end = m_model.getSimulationTime();
            m_events.clear();
            return;
        }

        if (m_simulatorUI != null) {
            if (current.getTime() >= m_model.getWarmupTime()) {
                m_simulatorUI.onEvent(current);
            }
        }

        Event next = current.advance();

        if (next != null) {
            m_events.add(next);
            m_end = Math.max(m_end, next.getTime());
        } else {
            m_end = Math.max(m_end, current.getTime());
        }
    }

    public void end() {
        m_model.endStatistics(m_end);

        if (m_simulatorUI != null) {
            m_simulatorUI.onEnd();
        }
    }

    public int getQueueLength(Widget widget) {
        if (!m_queues.containsKey(widget)) {
            return 0;
        } else {
            ListPriorityPair pair = (ListPriorityPair) m_queues.get(widget);
            return pair.list.size();
        }
    }

    public void enqueueEvent(Event event) {
        Widget widget = event.getWidget();

        List waiting;
        if (!m_queues.containsKey(widget)) {
            waiting = new ArrayList();
            m_queues.put(widget, new ListPriorityPair(waiting));
        } else {
            ListPriorityPair pair = (ListPriorityPair) m_queues.get(widget);
            waiting = pair.list;
        }

        waiting.add(event);

        if (m_simulatorUI != null) {
            if (event.getTime() > m_model.getWarmupTime()) {
                m_simulatorUI.onSuspendEvent(event);
            }
        }
    }

    public void dequeueEvent(QueueingWidget widget, double time) {
        ListPriorityPair pair = (ListPriorityPair) m_queues.get(widget);
        if (pair == null) {
            return;
        }

        Event event = widget.selectQueuedEvent(pair.list, pair.priority);
        if (event != null) {
            pair.priority = event.getTransactionPriority();
            pair.list.remove(event);
            m_events.add(new Event(event, time, Event.DEQUEUE));
        }
    }

    public void destroyEvent(Event event) {
        if (m_simulatorUI != null) {
            if (event.getTime() > m_model.getWarmupTime()) {
                m_simulatorUI.onDestroyEvent(event);
            }
        }
    }

    public int getTokens(WidgetResource resource) {
        Integer value;

        if (!m_resources.containsKey(resource)) {
            value = new Integer(resource.getNumberOfTokens());
            m_resources.put(resource, value);
        } else {
            value = (Integer) m_resources.get(resource);
        }

        return value.intValue();
    }

    public void setTokens(WidgetResource resource, int tokens) {
        m_resources.put(resource, new Integer(tokens));
    }

    public double getLastDepartureTime(WidgetSource source) {
        if (!m_sources.containsKey(source)) {
            return 0;
        } else {
            Double value = (Double) m_sources.get(source);
            return value.doubleValue();
        }
    }

    public void setLastDepartureTime(WidgetSource source, double time) {
        m_sources.put(source, new Double(time));
    }

    public void setSimulatorUI(SimulatorUI simulatorUI) {
        m_simulatorUI = simulatorUI;
    }

    public SimulatorUI getSimulatorUI() {
        return m_simulatorUI;
    }
}
