package edu.utexas.cs.surdules.pipes.model.widgets;

import edu.utexas.cs.surdules.pipes.model.Arc;
import edu.utexas.cs.surdules.pipes.model.Connector;
import edu.utexas.cs.surdules.pipes.model.Distributions;
import edu.utexas.cs.surdules.pipes.model.Model;
import edu.utexas.cs.surdules.pipes.model.QueueingWidget;
import edu.utexas.cs.surdules.pipes.model.Widget;
import edu.utexas.cs.surdules.pipes.model.forms.PropertiesItem;
import edu.utexas.cs.surdules.pipes.model.forms.PropertiesList;
import edu.utexas.cs.surdules.pipes.model.io.Visitor;
import edu.utexas.cs.surdules.pipes.model.simulate.Simulator;
import edu.utexas.cs.surdules.pipes.model.simulate.Event;

import java.util.Arrays;
import java.util.Iterator;
import java.util.List;

/**
 * @author <A HREF="mailto:razvan.surdulescu@post.harvard.edu">Razvan Surdulescu</A> (c) 2003
 * <BR>
 * You may use and distribute this software for free provided you include
 * this copyright notice. You may not sell this software, use my name for
 * publicity reasons or modify the code without permission from me.
 */
public class WidgetService extends Widget implements QueueingWidget {
    public static final String TIMERULE_FIFO = "FIFO";
    public static final String TIMERULE_LIFO = "LIFO";

    private static final String[] s_timeRuleNames = new String[]{
        TIMERULE_FIFO, TIMERULE_LIFO
    };

    public static final String QUEUEINGPRIORITY_NOPRIORITY = "No Priority";
    public static final String QUEUEINGPRIORITY_NEXTPRIORITY = "Next Priority";
    public static final String QUEUEINGPRIORITY_HIGHESTPRIORITY = "Highest Priority";

    private static final String[] s_queueingPriorityNames = new String[]{
        QUEUEINGPRIORITY_NOPRIORITY, QUEUEINGPRIORITY_NEXTPRIORITY, QUEUEINGPRIORITY_HIGHESTPRIORITY
    };

    private String m_distributionName;
    private double m_distributionValueA, m_distributionValueB;
    private String m_timeRuleName;
    private String m_queueingPriorityName;

    public WidgetService(Model model) {
        this(model, 0, 0);
    }

    public WidgetService(Model model, int x, int y) {
        super(model, x, y);

        m_distributionName = Distributions.getDistributionNames()[0];
        m_timeRuleName = s_timeRuleNames[0];
        m_queueingPriorityName = s_queueingPriorityNames[0];
    }

    protected void createConnectors() {
        Connector incoming = new Connector(this, Connector.INCOMING, 10, Connector.WEST, 0);
        m_connectors.add(incoming);

        Connector outgoing = new Connector(this, Connector.OUTGOING, 10, Connector.EAST, 1);
        m_connectors.add(outgoing);
    }

    protected String getDefaultName() {
        return "Service";
    }

    public PropertiesList getPropertiesList() {
        PropertiesList propertiesList = super.getPropertiesList();
        propertiesList.addItem(new PropertiesItem("Service Time Distribution", String.class, m_distributionName,
                Arrays.asList(Distributions.getDistributionNames())));
        propertiesList.addItem(new PropertiesItem("Distribution Parameter a", Double.class, new Double(m_distributionValueA)));
        propertiesList.addItem(new PropertiesItem("Distribution Parameter b", Double.class, new Double(m_distributionValueB)));
        propertiesList.addItem(new PropertiesItem("Time Rule", String.class, m_timeRuleName,
                Arrays.asList(s_timeRuleNames)));
        propertiesList.addItem(new PropertiesItem("Queueing Priority", String.class, m_queueingPriorityName,
                Arrays.asList(s_queueingPriorityNames)));
        return propertiesList;
    }

    public void setPropertiesList(PropertiesList propertiesList) {
        super.setPropertiesList(propertiesList);

        String distributionName = (String) propertiesList.getItem("Service Time Distribution").getValue();
        double distributionValueA = ((Double) propertiesList.getItem("Distribution Parameter a").getValue()).doubleValue();
        double distributionValueB = ((Double) propertiesList.getItem("Distribution Parameter b").getValue()).doubleValue();

        Distributions distributions = new Distributions();
        // This will throw an IllegalArgumentException if the distribution
        // parameters are incorrect.
        distributions.computeNamedDistribution(distributionName, distributionValueA, distributionValueB);

        m_distributionName = distributionName;
        m_distributionValueA = distributionValueA;
        m_distributionValueB = distributionValueB;

        m_timeRuleName = (String) propertiesList.getItem("Time Rule").getValue();
        m_queueingPriorityName = (String) propertiesList.getItem("Queueing Priority").getValue();
    }

    public String getDistributionName() {
        return m_distributionName;
    }

    public void setDistributionName(String distributionName) {
        m_distributionName = distributionName;
    }

    public double getDistributionValueA() {
        return m_distributionValueA;
    }

    public void setDistributionValueA(double distributionValueA) {
        m_distributionValueA = distributionValueA;
    }

    public double getDistributionValueB() {
        return m_distributionValueB;
    }

    public void setDistributionValueB(double distributionValueB) {
        m_distributionValueB = distributionValueB;
    }

    public String getTimeRuleName() {
        return m_timeRuleName;
    }

    public void setTimeRuleName(String timeRuleName) {
        m_timeRuleName = timeRuleName;
    }

    public String getQueueingPriorityName() {
        return m_queueingPriorityName;
    }

    public void setQueueingPriorityName(String queueingPriorityName) {
        m_queueingPriorityName = queueingPriorityName;
    }

    public void accept(Visitor visitor) {
        visitor.visitWidgetService(this);
        super.accept(visitor);
    }

    public Event advance(Event event, Simulator context) {
        if (event.getWidget() != this) {
            throw new IllegalArgumentException("Cannot process Event belonging to another Widget");
        }

        if (event.getState() == Event.ENTER || event.getState() == Event.DEQUEUE) {
            if (event.getState() == Event.ENTER) {
                m_statistics.onArrive(event);
            } else {
                m_statistics.onDequeue(event);
            }

            if (context.getQueueLength(this) > 0) {
                context.enqueueEvent(event);
                m_statistics.onEnqueue(event);
                return null;
            } else {
                double work = context.getDistributions().computeNamedDistribution(
                        getDistributionName(),
                        getDistributionValueA(),
                        getDistributionValueB());
                m_statistics.onWork(event, work);

                double time = event.getTime() + work;

                context.dequeueEvent(this, time);

                return new Event(event, time, Event.LEAVE);
            }
        } else if (event.getState() == Event.LEAVE) {
            Arc arc = selectOutgoingArc(event.getTransactionCategory(), context);

            if (arc == null) {
                m_statistics.onDestroy(event);
                return null;
            } else {
                m_statistics.onLeave(event);
                return new Event(event, arc.getEnd().getWidget(), Event.ENTER);
            }
        } else {
            throw new IllegalArgumentException("Unrecognized event state: " + event.getState());
        }
    }

    public Event selectQueuedEvent(List events, int priority) {
        if (events == null) {
            return null;
        }

        Event result = null;

        if (m_queueingPriorityName.equals(QUEUEINGPRIORITY_NOPRIORITY)) {
            if (m_timeRuleName.equals(TIMERULE_FIFO)) {
                for (Iterator iter = events.iterator(); iter.hasNext();) {
                    Event current = (Event) iter.next();
                    if (result == null) {
                        result = current;
                    } else if (current.getTime() < result.getTime()) {
                        result = current;
                    }
                }
            } else if (m_timeRuleName.equals(TIMERULE_LIFO)) {
                for (Iterator iter = events.iterator(); iter.hasNext();) {
                    Event current = (Event) iter.next();
                    if (result == null) {
                        result = current;
                    } else if (current.getTime() > result.getTime()) {
                        result = current;
                    }
                }
            } else {
                throw new IllegalArgumentException("Unrecognized time rule name: " + m_timeRuleName);
            }
        } else if (m_queueingPriorityName.equals(QUEUEINGPRIORITY_NEXTPRIORITY)) {
            if (priority == 0) {
                priority = Integer.MAX_VALUE;
            }

            Event max = null;

            for (Iterator iter = events.iterator(); iter.hasNext();) {
                Event current = (Event) iter.next();
                if (current.getTransactionPriority() >= priority) {
                    if (max == null) {
                        max = current;
                    } else if (current.getTransactionPriority() > max.getTransactionPriority()) {
                        max = current;
                    }
                } else if (result == null) {
                    result = current;
                } else if (current.getTransactionPriority() > result.getTransactionPriority()) {
                    result = current;
                }
            }

            if (result == null) {
                return max;
            } else {
                return result;
            }
        } else if (m_queueingPriorityName.equals(QUEUEINGPRIORITY_HIGHESTPRIORITY)) {
            for (Iterator iter = events.iterator(); iter.hasNext();) {
                Event current = (Event) iter.next();
                if (result == null) {
                    result = current;
                } else if (current.getTransactionPriority() > result.getTransactionPriority()) {
                    result = current;
                }
            }
        } else {
            throw new IllegalArgumentException("Unrecognized queueing priority name: " + m_queueingPriorityName);
        }

        return result;
    }
}
