package edu.utexas.cs.surdules.pipes.model.widgets;

import edu.utexas.cs.surdules.pipes.model.Arc;
import edu.utexas.cs.surdules.pipes.model.Connector;
import edu.utexas.cs.surdules.pipes.model.Distributions;
import edu.utexas.cs.surdules.pipes.model.Model;
import edu.utexas.cs.surdules.pipes.model.Widget;
import edu.utexas.cs.surdules.pipes.model.forms.PropertiesItem;
import edu.utexas.cs.surdules.pipes.model.forms.PropertiesList;
import edu.utexas.cs.surdules.pipes.model.io.Visitor;
import edu.utexas.cs.surdules.pipes.model.simulate.Simulator;
import edu.utexas.cs.surdules.pipes.model.simulate.Event;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;

/**
 * @author <A HREF="mailto:razvan.surdulescu@post.harvard.edu">Razvan Surdulescu</A> (c) 2003
 * <BR>
 * You may use and distribute this software for free provided you include
 * this copyright notice. You may not sell this software, use my name for
 * publicity reasons or modify the code without permission from me.
 */
public class WidgetSource extends Widget {
    private String m_distributionName;
    private double m_distributionValueA, m_distributionValueB;
    private int m_numberOfTransactions;
    private String m_transactionCategory;
    private int m_transactionPriority;

    public WidgetSource(Model model) {
        this(model, 0, 0);
    }

    public WidgetSource(Model model, int x, int y) {
        super(model, x, y);

        m_distributionName = Distributions.getDistributionNames()[0];
        m_numberOfTransactions = 1;
    }

    protected void createConnectors() {
        Connector outgoing = new Connector(this, Connector.OUTGOING, 10, Connector.EAST, 0);
        m_connectors.add(outgoing);
    }

    protected String getDefaultName() {
        return "Source";
    }

    public PropertiesList getPropertiesList() {
        PropertiesList propertiesList = super.getPropertiesList();
        propertiesList.addItem(new PropertiesItem("Interdeparture Time Distribution", String.class, m_distributionName,
                Arrays.asList(Distributions.getDistributionNames())));
        propertiesList.addItem(new PropertiesItem("Distribution Parameter a", Double.class, new Double(m_distributionValueA)));
        propertiesList.addItem(new PropertiesItem("Distribution Parameter b", Double.class, new Double(m_distributionValueB)));
        propertiesList.addItem(new PropertiesItem("Number of Transactions", Integer.class, new Integer(m_numberOfTransactions)));

        List categories = new ArrayList();
        categories.add(PropertiesItem.NONE);
        for (Iterator iter = m_model.getWidgets(WidgetSource.class).iterator(); iter.hasNext();) {
            WidgetSource source = (WidgetSource) iter.next();
            String category = source.getTransactionCategory();
            if (category != null) {
                categories.add(category);
            }
        }
        categories.add(PropertiesItem.CREATENEW);

        propertiesList.addItem(new PropertiesItem("Transaction Category", String.class,
                (m_transactionCategory == null? PropertiesItem.NONE : m_transactionCategory), categories));

        propertiesList.addItem(new PropertiesItem("Transaction Priority", Integer.class, new Integer(m_transactionPriority)));

        return propertiesList;
    }

    public void setPropertiesList(PropertiesList propertiesList) {
        super.setPropertiesList(propertiesList);

        String distributionName = (String) propertiesList.getItem("Interdeparture Time Distribution").getValue();
        double distributionValueA = ((Double) propertiesList.getItem("Distribution Parameter a").getValue()).doubleValue();
        double distributionValueB = ((Double) propertiesList.getItem("Distribution Parameter b").getValue()).doubleValue();

        Distributions distributions = new Distributions();
        // This will throw an IllegalArgumentException if the distribution
        // parameters are incorrect.
        distributions.computeNamedDistribution(distributionName, distributionValueA, distributionValueB);

        m_distributionName = distributionName;
        m_distributionValueA = distributionValueA;
        m_distributionValueB = distributionValueB;

        setNumberOfTransactions(((Integer) propertiesList.getItem("Number of Transactions").getValue()).intValue());

        String category = (String) propertiesList.getItem("Transaction Category").getValue();
        if (category.equals(PropertiesItem.NONE)) {
            m_transactionCategory = null;
        } else {
            m_transactionCategory = category;
        }

        setTransactionPriority(((Integer) propertiesList.getItem("Transaction Priority").getValue()).intValue());
    }

    public String getDistributionName() {
        return m_distributionName;
    }

    public void setDistributionName(String distributionName) {
        m_distributionName = distributionName;
    }

    public double getDistributionValueA() {
        return m_distributionValueA;
    }

    public void setDistributionValueA(double distributionValueA) {
        m_distributionValueA = distributionValueA;
    }

    public double getDistributionValueB() {
        return m_distributionValueB;
    }

    public void setDistributionValueB(double distributionValueB) {
        m_distributionValueB = distributionValueB;
    }

    public int getNumberOfTransactions() {
        return m_numberOfTransactions;
    }

    public void setNumberOfTransactions(int numberOfTransactions) {
        if (numberOfTransactions < 1) {
            throw new IllegalArgumentException("Number of transactions must be at least 1: " + numberOfTransactions);
        }
        m_numberOfTransactions = numberOfTransactions;
    }

    public String getTransactionCategory() {
        return m_transactionCategory;
    }

    public void setTransactionCategory(String category) {
        m_transactionCategory = category;
    }

    public int getTransactionPriority() {
        return m_transactionPriority;
    }

    public void setTransactionPriority(int priority) {
        if (priority < 0) {
            throw new IllegalArgumentException("Event priority cannot be negative: " + priority);
        }

        m_transactionPriority = priority;
    }

    public void accept(Visitor visitor) {
        visitor.visitWidgetSource(this);
        super.accept(visitor);
    }

    public Event advance(Event event, Simulator context) {
        if (event.getWidget() != this) {
            throw new IllegalArgumentException("Cannot process Event belonging to another Widget");
        }

        if (event.getState() == Event.ENTER) {
            m_statistics.onCreate(event);

            double time = context.getLastDepartureTime(this) + event.getTime() +
                    context.getDistributions().computeNamedDistribution(
                            getDistributionName(),
                            getDistributionValueA(),
                            getDistributionValueB());
            context.setLastDepartureTime(this, time);
            return new Event(event, time, Event.LEAVE);
        } else if (event.getState() == Event.DEQUEUE) {
            throw new IllegalArgumentException("Cannot handle dequeued event on a widget without a queue: " + event);
        } else if (event.getState() == Event.LEAVE) {
            Arc arc = selectOutgoingArc(event.getTransactionCategory(), context);

            if (arc == null) {
                m_statistics.onDestroy(event);
                return null;
            } else {
                m_statistics.onLeave(event);
                return new Event(event, arc.getEnd().getWidget(), Event.ENTER);
            }
        } else {
            throw new IllegalArgumentException("Unrecognized event state: " + event.getState());
        }
    }
}
